#!/usr/bin/perl -w
#
# File: GenerateLMOntologyData.pl
#
# Author: Manish Sud
#
# Copyright (C) 2006-2012. The Regents of the University of California. All Rights Reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the LIPID MAPS (TM) consortium nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# LIPID MAPS (TM) Tools software package is developed by Manish Sud and Eoin Fahy
# at San Diego Supercomputer Center at the University of California, San Diego as
#  a part of LIPID Metabolites and Pathways Strategy consortium efforts.
#
# The LIPID MAPS (TM) consortium is a multi-institutional multi-year effort
# involved in identification and characterization of existing and novel lipids,
# quantifications of changes in their metabolites, and development of biochemical
# pathways and interaction network maps involving lipids. Please acknowledge use
# of the LIPID MAPS (TM) Tools in your work by citing appropriate literature
# references. For further details, please visit www.lipidmaps.org.
#
# Support for this effort is provided by  by National Institutes of Health (NIH)
# and National Institute of General Medical Sciences (NIGMS) Glue Grant
# NIH/NIGMS Grant 1 U54 GM69338.
#

# Make sure MayaChemTools package is already installed...
if (exists $ENV{MAYACHEMTOOLS} && $ENV{MAYACHEMTOOLS}) {
  use lib "$ENV{MAYACHEMTOOLS}/lib";
}
else {
  die "The LIPID MAPS ontology data generation script couldn't be started: It uses Perl modules from MayaChemTools package to generate ontology data and it's not available on your system through MAYACHEMTOOLS environment variable. Please:\n\no Download and install MayaChemTools package available at: www.MayaChemTools.org\no Set MAYACHEMTOOLS environment variable point to <YOUR MAYACHEMTOOLS ROOT DIR>\no And try again\n\n";
}

use strict;
use FindBin;
use Getopt::Long;
use File::Basename;
use Text::ParseWords;
use Cwd;
use Benchmark;
use TextUtil;
use FileUtil;
use SDFileUtil;
use MoleculeFileIO;

my($ScriptName, $ScriptUsage, %Options, $StartTime, $EndTime, $TotalTime);

# Autoflush STDOUT
$| = 1;

# Starting message...
$ScriptName = basename $0;
print "\n$ScriptName:Starting...\n\n";
$StartTime = new Benchmark;

# Get the options and setup script...
SetupScriptUsage();
if ($Options{help} || @ARGV < 1) {
  die GetUsageFromPod("$FindBin::Bin/$ScriptName");
}

my(@SDFilesList);

@SDFilesList = ExpandFileNames(\@ARGV, "sdf sd");

print "Processing options...\n";
my($Mode, $CompoundIDMode, $CompoundIDLabel, $CompoundID, $Fingerprints,  $Output, $SDOutput, $TextOutput, $OutDelim, $OutQuote, $OverwriteFiles, $OutFileRoot, $FingerprintsString, @OntologyDataFieldNames, %OntologyDataFieldNameToIndexMap);
ProcessOptions();

# Collect information about SD files...
my(%SDFilesInfo);
print "Checking input SD file(s)...\n";
RetrieveSDFilesInfo();

# Process input SD files..
my($FileIndex, $SDFile, $FileProcessingMsg, $CmpdCount, $IgnoredCmpdCount, $ProcessLMIDValue, $CompoundIDDataFieldValue);

$FileProcessingMsg = "Processing file";
if (@SDFilesList > 1) {
  print "Processing SD files...\n";
  $FileProcessingMsg = "\n$FileProcessingMsg";
}

for $FileIndex (0 .. $#SDFilesList) {
  if ($SDFilesInfo{FileOkay}[$FileIndex]) {
    $SDFile = $SDFilesList[$FileIndex];
    print "$FileProcessingMsg $SDFile...\n";
    ProcessSDFile($FileIndex);
  }
}
print "$ScriptName:Done...\n\n";

$EndTime = new Benchmark;
$TotalTime = timediff ($EndTime, $StartTime);
print "Total time: ", timestr($TotalTime), "\n";

###############################################################################

# Generate ontology data for a SD file...
sub ProcessSDFile {
  my($FileIndex) = @_;
  my($SDFile, $MoleculeFileIO, $Molecule, $LipidOntologyDataRef, $NewSDFile, $NewTextFile, $NewSDFileRef, $NewTextFileRef);

  $SDFile = $SDFilesList[$FileIndex];

  # Setup output files...
  $NewSDFileRef = '';
  $NewTextFileRef = '';
  ($NewSDFileRef, $NewTextFileRef) = SetupAndOpenOutputFiles($FileIndex);

  $MoleculeFileIO = new MoleculeFileIO('Name' => $SDFile);
  $MoleculeFileIO->Open();

  $CmpdCount = 0;
  $IgnoredCmpdCount = 0;

  COMPOUND: while ($Molecule = $MoleculeFileIO->ReadMolecule()) {
    $CmpdCount++;

    # Validate compound data before ontology calculation...
    if (!ValidateCompoundData($Molecule)) {
      next COMPOUND;
    }

    # Generate LM ontology data...
    print "\nGenerateLipidOntologyData for compound record number $CmpdCount...\n";
    $LipidOntologyDataRef = GenerateLipidOntologyData($Molecule);

    # Warn and ignore compound...
    if (!$LipidOntologyDataRef) {
      $IgnoredCmpdCount++;
      ProcessIgnoredCompound('OntologyCalcFailed', $Molecule);
      next COMPOUND;
    }

    # Adjust ontology values for specific lipid classes...
    if ($ProcessLMIDValue && $CompoundIDDataFieldValue) {
      AdjustOntologyDataValues($CompoundIDDataFieldValue, $Molecule, $LipidOntologyDataRef);
    }

    # Generate fingerprint descriptor...
    if ($Fingerprints) {
      GenerateLipidOntologyDataForFingerprints($LipidOntologyDataRef);
    }

    # Generate output files...
    WriteDataToOutputFiles($FileIndex, $CmpdCount, $Molecule, $LipidOntologyDataRef, $NewSDFileRef, $NewTextFileRef);
  }
  $MoleculeFileIO->Close();

  if ($SDOutput) {
    close $NewSDFileRef;
  }
  if ($TextOutput) {
    close $NewTextFileRef;
  }
  WriteOntologyGenerationSummaryStatistics($CmpdCount, $IgnoredCmpdCount);
}

# Open output files...
#
sub SetupAndOpenOutputFiles {
  my($FileIndex) = @_;
  my($NewSDFile, $NewTextFile, $NewSDFileRef, $NewTextFileRef);

  $NewSDFileRef = '';
  $NewTextFileRef = '';

  if ($SDOutput) {
    $NewSDFile = $SDFilesInfo{SDOutFileNames}[$FileIndex];
    print "Generating SD file $NewSDFile...\n";
    open NEWSDFILE, ">$NewSDFile" or die "Error: Couldn't open $NewSDFile: $! \n";
    $NewSDFileRef = \*NEWSDFILE;
  }

  if ($TextOutput) {
    my($Line, @LineWords);

    $NewTextFile = $SDFilesInfo{TextOutFileNames}[$FileIndex];
    print "Generating text file $NewTextFile...\n";
    open NEWTEXTFILE, ">$NewTextFile" or die "Error: Couldn't open $NewTextFile: $! \n";

    @LineWords = ();

    # Compound ID label...
    push @LineWords, $CompoundIDLabel;

    # Ontology data labels...
    push @LineWords, @OntologyDataFieldNames;

    # Fingerprint label...
    if ($Fingerprints) {
      push @LineWords, "FINGERPRINT";
    }

    $Line = JoinWords(\@LineWords, $OutDelim, $OutQuote);
    print NEWTEXTFILE "$Line\n";

    $NewTextFileRef = \*NEWTEXTFILE;
  }
  return ($NewSDFileRef, $NewTextFileRef);
}

# Write out lipid ontology data...
#
sub WriteDataToOutputFiles {
  my($FileIndex, $CmpdCount, $Molecule, $LipidOntologyDataRef, $NewSDFileRef, $NewTextFileRef) = @_;


  if ($SDOutput) {
    my($Index, $DataFieldName, $DataFieldValue);

    # Retrieve input compound string used to create molecule and write it out
    # without last line containing a delimiter...
    #
    my($CmpdString);
    $CmpdString = $Molecule->GetInputMoleculeString();
    $CmpdString =~ s/\$\$\$\$$//;
    print $NewSDFileRef "$CmpdString";

    for $Index (0 .. $#OntologyDataFieldNames) {
      $DataFieldName = $OntologyDataFieldNames[$Index];
      $DataFieldValue = $LipidOntologyDataRef->[$Index];

      print $NewSDFileRef  ">  <$DataFieldName>\n$DataFieldValue\n\n";
    }
    if ($Fingerprints) {
      print $NewSDFileRef  ">  <FINGERPRINT>\n$FingerprintsString\n\n";
    }

    # Write out delimiter...
    print $NewSDFileRef "\$\$\$\$\n";
  }

  if ($TextOutput) {
    my($Line, $CmpdID, $DataFieldLabelAndValuesRef, @LineWords);

    @LineWords = ();

    $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
    $CmpdID = SetupCmpdIDForTextFiles($Molecule, $DataFieldLabelAndValuesRef);
    push @LineWords, $CmpdID;

    push @LineWords, @{$LipidOntologyDataRef};
    if ($Fingerprints) {
      push @LineWords, $FingerprintsString;
    }

    $Line = JoinWords(\@LineWords, $OutDelim, $OutQuote);
    print $NewTextFileRef "$Line\n";
  }

}

# Process compound being ignored due to problems in ontology geneation...
#
sub ProcessIgnoredCompound {
  my($Mode, $Molecule) = @_;
  my($CmpdID, $DataFieldLabelAndValuesRef);

  $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
  $CmpdID = SetupCmpdIDForTextFiles($Molecule, $DataFieldLabelAndValuesRef);

  MODE: {
    if ($Mode =~ /^MissingCompoundIDField$/i) {
      warn "Warning: Ignoring compound record number $CmpdCount with ID $CmpdID: Specified data field name, $CompoundID, is missing...\n\n";
      next MODE;
    }

    if ($Mode =~ /^ContainsNonElementalData$/i) {
      warn "Warning: Ignoring compound record number $CmpdCount with ID $CmpdID: Compound contains non-elemental atom symbol(s)...\n\n";
      next MODE;
    }

    if ($Mode =~ /^ContainsNoElementalData$/i) {
      warn "Warning: Ignoring compound record number $CmpdCount with ID $CmpdID: Compound contains no element atom symbol(s)...\n\n";
      next MODE;
    }


    if ($Mode =~ /^OntologyCalcFailed$/i) {
      warn "Warning: Ignoring compound record number $CmpdCount with ID $CmpdID: ontology generation didn't succeed...\n\n";
      next MODE;
    }
    warn "Warning: Ignoring compound record number $CmpdCount with ID $CmpdID...\n\n";
  }
}

# Setup compound ID...
#
sub SetupCmpdIDForTextFiles {
  my($Molecule, $DataFieldLabelAndValuesRef) = @_;
  my($CmpdID);

  $CmpdID = '';
  if ($CompoundIDMode =~ /^MolNameOrLabelPrefix$/i) {
    my($MolName);
    $MolName = $Molecule->GetName();
    $CmpdID = $MolName ? $MolName : "${CompoundID}${CmpdCount}";
  }
  elsif ($CompoundIDMode =~ /^LabelPrefix$/i) {
    $CmpdID = "${CompoundID}${CmpdCount}";
  }
  elsif ($CompoundIDMode =~ /^DataField$/i) {
    my($SpecifiedDataField);
    $SpecifiedDataField = $CompoundID;
    $CmpdID = exists $DataFieldLabelAndValuesRef->{$SpecifiedDataField} ? $DataFieldLabelAndValuesRef->{$SpecifiedDataField} : '';
  }
  elsif ($CompoundIDMode =~ /^MolName$/i) {
    $CmpdID = $Molecule->GetName();
  }

  if (IsEmpty($CmpdID)) {
    $CmpdID = "${CompoundID}${CmpdCount}";
  }
  return $CmpdID;
}

# Validate compound element data and make sure appropriate data field is present...
#
sub ValidateCompoundData {
  my($Molecule)  = @_;
  my($DataFieldLabelAndValuesRef, $SpecifiedDataField);

  # Check for any non-elemental data...
  # Notes:
  #    . Any isotopes: D and so on
  #    . Functional groups as atomic symbols
  #
  my($Atom, $AtomicNumber, $AtomsCount);

  $AtomsCount = 0;
  for $Atom ($Molecule->GetAtoms()) {
    $AtomsCount++;
    $AtomicNumber = $Atom->GetAtomicNumber();
    if (!$AtomicNumber) {
      $IgnoredCmpdCount++;
      ProcessIgnoredCompound('ContainsNonElementalData', $Molecule);
      return 0;
    }
  }

  if (!$AtomsCount) {
    $IgnoredCmpdCount++;
    ProcessIgnoredCompound('ContainsNoElementalData', $Molecule);
    return 0;
  }

  # Check specified data field...
  $CompoundIDDataFieldValue = undef;
  if ($CompoundIDMode =~ /^DataField$/) {
    $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
    if (exists $DataFieldLabelAndValuesRef->{$CompoundID}) {
      $CompoundIDDataFieldValue = $DataFieldLabelAndValuesRef->{$CompoundID};
    }
    else {
      # Warn and ignore compound...
      $IgnoredCmpdCount++;
      ProcessIgnoredCompound('MissingCompoundIDField', $Molecule);
      return 0;
    }
  }

  return 1;
}

# Write out compounds ontology generation summary statistics...
#
sub WriteOntologyGenerationSummaryStatistics {
  my($CmpdCount, $IgnoredCmpdCount) = @_;
  my($ProcessedCmpdCount);

  $ProcessedCmpdCount = $CmpdCount - $IgnoredCmpdCount;

  print "\nNumber of compounds: $CmpdCount\n";
  print "Number of compounds processed successfully during ontology generation: $ProcessedCmpdCount\n";
  print "Number of compounds ignored during ontology generation: $IgnoredCmpdCount\n";
}

# Adjust ontology data values for specific lipid categories...
#
sub AdjustOntologyDataValues {
  my($LMID, $Molecule, $LipidOntologyDataRef) = @_;
  my($Nothing);

  LMID: {

    if ($LMID =~ /^LMGL/i) {
      AdjustOntologyDataValuesForGlycerolipids($LMID, $Molecule, $LipidOntologyDataRef);
      last LMID;
    }

    if ($LMID =~ /^LMGP/i) {
      AdjustOntologyDataValuesForGlycerophospholipids($LMID, $Molecule, $LipidOntologyDataRef);
      last LMID;
    }

    if ($LMID =~ /^LMSP/i) {
      AdjustOntologyDataValuesForSphingolipids($LMID, $Molecule, $LipidOntologyDataRef);
      last LMID;
    }

    $Nothing = 1;
  }
}

# Adjust ontology data values for glycerolipids...
#
sub AdjustOntologyDataValuesForGlycerolipids {
  my($LMID, $Molecule, $LipidOntologyDataRef) = @_;
  my($Name, $DataFieldLabelAndValuesRef);

  $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
  $Name = exists $DataFieldLabelAndValuesRef->{NAME} ? $DataFieldLabelAndValuesRef->{NAME} :undef;

  # Does Name contain valid chain length appreviations?
  if ($Name && IsValidNameAbbreviation($Name)) {
    AdjustCARBONSOntologyDataValueUsingChainLengthAbbreviations($LMID, $Name, $Molecule, $LipidOntologyDataRef);
  }
}

# Adjust ontology data values for glycerophospholipids...
#
sub AdjustOntologyDataValuesForGlycerophospholipids {
  my($LMID, $Molecule, $LipidOntologyDataRef) = @_;
  my($Name, $DataFieldLabelAndValuesRef);

  $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
  $Name = exists $DataFieldLabelAndValuesRef->{NAME} ? $DataFieldLabelAndValuesRef->{NAME} :undef;

  # Does Name contain valid chain length appreviations?
  if ($Name && IsValidNameAbbreviation($Name)) {
    AdjustCARBONSOntologyDataValueUsingChainLengthAbbreviations($LMID, $Name, $Molecule, $LipidOntologyDataRef);
  }
}

# Adjust ontology data values for sphingolipids...
#
sub AdjustOntologyDataValuesForSphingolipids {
  my($LMID, $Molecule, $LipidOntologyDataRef) = @_;
  my($Name, $DataFieldLabelAndValuesRef);

  $DataFieldLabelAndValuesRef = $Molecule->GetDataFieldLabelAndValues();
  $Name = exists $DataFieldLabelAndValuesRef->{NAME} ? $DataFieldLabelAndValuesRef->{NAME} :undef;

  # Does Name contain valid chain length appreviations?
  if ($Name && IsValidNameAbbreviation($Name)) {
    AdjustCARBONSOntologyDataValueUsingChainLengthAbbreviations($LMID, $Name, $Molecule, $LipidOntologyDataRef);
  }
}

# Adjust ontology data value for CARBONS
#
# Notes:
#    . The number of Carbons in ontology data correspond to the total number of  Carbons.
#      Subtract ME, ET and PR Carbons to map Carbon count to chain atom acount.
#
sub AdjustCARBONSOntologyDataValuesForAlkylSubstituents {
  my($LMID, $Molecule, $LipidOntologyDataRef) = @_;
  my($CARBONSIndex);

  if (!exists $OntologyDataFieldNameToIndexMap{CARBONS}) {
    return;
  }
  $CARBONSIndex = $OntologyDataFieldNameToIndexMap{CARBONS};
  if (!$LipidOntologyDataRef->[$CARBONSIndex]) {
    return;
  }

  # Adjust for ME, ET and PR
  my($Index, $Name, $Value, $NameIndex, $Multiplier, $TotalAdjustment, @DataFieldNames, @AdjustmentMultiplier);

  @DataFieldNames = ('METHYLS', 'ET', 'PR');
  @AdjustmentMultiplier = (1, 2, 3);
  $TotalAdjustment = 0;

  INDEX: for $Index (0 .. $#DataFieldNames) {
    $Name = $DataFieldNames[$Index];
    $Multiplier = $AdjustmentMultiplier[$Index];
    if (!exists $OntologyDataFieldNameToIndexMap{$Name}) {
      next INDEX;
    }

    $NameIndex = $OntologyDataFieldNameToIndexMap{$Name};
    if (!$LipidOntologyDataRef->[$NameIndex]) {
      next INDEX;
    }
    $Value = $LipidOntologyDataRef->[$NameIndex];

    $TotalAdjustment += $Value * $Multiplier;
  }

  # Any adjustment to apply...
  if ($TotalAdjustment) {
    $LipidOntologyDataRef->[$CARBONSIndex] = $LipidOntologyDataRef->[$CARBONSIndex] - $TotalAdjustment;
    print "Adjusting calculated CARBONS value $LipidOntologyDataRef->[$CARBONSIndex] by subtracting $TotalAdjustment corresponding to sum of Carbons in METHYLS, ET and PR alkyl substituents present in the compound...\n";
  }
}

# Adjust CARBONS ontology data using abbreviation chain lengths...
#
sub AdjustCARBONSOntologyDataValueUsingChainLengthAbbreviations {
  my($LMID, $Name, $Molecule, $LipidOntologyDataRef) = @_;
  my($CARBONSIndex, $ChainCARBONSCount, $LMType, $ChainsAbbrev, $AbbrevModifier, $ChainAbbrev, $ChainLength, $DoubleBondCount, $ChanLengthsOkay, @ChainsAbbrevList);

  if (!exists $OntologyDataFieldNameToIndexMap{CARBONS}) {
    return;
  }
  $CARBONSIndex = $OntologyDataFieldNameToIndexMap{CARBONS};
  if (!$LipidOntologyDataRef->[$CARBONSIndex]) {
    return;
  }

  $ChanLengthsOkay = 1;
  $ChainCARBONSCount = 0;
  ($LMType, $ChainsAbbrev, $AbbrevModifier) = ParseNameAbbrev($Name);
  @ChainsAbbrevList = split /\//, $ChainsAbbrev;

  CHAINABBREV: for $ChainAbbrev (@ChainsAbbrevList) {
    ($ChainLength, $DoubleBondCount) = ParseChainAbbrev($ChainAbbrev);
    if (IsEmpty($ChainLength)) {
      next CHAINABBREV;
    }
    # Make sure chain length is numeric; otherwise, return empty chain length value...
    if (!IsInteger($ChainLength)) {
      $ChanLengthsOkay = 0;
      print "Warning: Chain length abbreviation, $ChainLength, is not an integer; it cann't be used to adjust CARBONS count...\n";
      last CHAINABBREV;
    }
    $ChainCARBONSCount += $ChainLength;
  }

  if ($ChanLengthsOkay && $ChainCARBONSCount) {
    print "Replacing calculated CARBONS value $LipidOntologyDataRef->[$CARBONSIndex] by $ChainCARBONSCount corresponding to sum of chain lengths in $Name...\n";
    $LipidOntologyDataRef->[$CARBONSIndex] = $ChainCARBONSCount;
  }
}

# Parse name abbreviation...
sub ParseNameAbbrev {
  my($Abbrev) = @_;
  my($LMType, $ChainsAbbrev, $AbbrevModifier);

  if ($Abbrev =~ /\]$/) {
    ($LMType, $ChainsAbbrev, $AbbrevModifier) = $Abbrev =~ /^(.+?)\((.+?)\)\[(.+?)$/;
    $AbbrevModifier = '[' . $AbbrevModifier;
  }
  else {
    ($LMType, $ChainsAbbrev) = $Abbrev =~ /^(.+?)\((.+?)\)$/;
    $AbbrevModifier = '';
  }
  return ($LMType, $ChainsAbbrev, $AbbrevModifier);
}

# Does Name correspond to valid chain length based abbreviations?
#
# For example:
#    . MG(16:0/), MG(0:0/20:4(5Z,8Z,11Z,14Z)/0:0)
#    . DGTS(16:0/16:0)
#    . TG(15:0/17:1(9Z)/15:0)
#    . And so on...
#
sub IsValidNameAbbreviation {
  my($Name) = @_;

  # Number of '(' and ')' parenthesis must match...
  my(@LeftParenthesis) = $Name =~ /\(/g;
  my(@RightParenthesis) = $Name =~ /\)/g;
  if ($#LeftParenthesis != $#RightParenthesis) {
    return 0;
  }

  @LeftParenthesis = $Name =~ /\[/g;
  @RightParenthesis = $Name =~ /\]/g;
  if ($#LeftParenthesis != $#RightParenthesis) {
    return 0;
  }

  my($LMType, $ChainsAbbrev, $AbbrevModifier);
  ($LMType, $ChainsAbbrev, $AbbrevModifier) = ParseNameAbbrev($Name);

  return (defined($LMType) && defined($ChainsAbbrev)) ? 1 : 0;
}

# Parse chain abbreviation to retrieve chain length, double bonds count and their geometry...
# Format:
#   ChainLen:BondCount(BondGeometry)(Substituents){Ring}
#
#   . Optional fields: Substituents and Rings; BondGeometry when BondCount = 0
#   . Stereo chemistry for substituents doesn't contain any numbers
#   . Substituent position must be > 0 and <= chain length (< chain length for epoxy)
#
#  e.g.
#   18:1(9Z)(6Ke,15OH[S]){8,9}
#   18:1(9Z)(6Ke,12OH[S],14Ep[R S]){8,9}
#   18:1(9Z)
#   18:0(9Ke,15OH)
#   and so on...
#
sub ParseChainAbbrev {
  my($ChainAbbrev) = @_;
  my($ChainLength, $DoubleBondCount, $DoubleBondGeometry, $Substituents, $Rings, $ChainSpec, $LeftOverChainSpec);

  $ChainLength = ''; $DoubleBondCount = ''; $DoubleBondGeometry = ''; $Substituents = ''; $Rings = '';
  $ChainSpec = $ChainAbbrev; $LeftOverChainSpec = '';
  ($ChainLength, $DoubleBondCount, $LeftOverChainSpec) = $ChainSpec =~ /^(.+?)\:(.+?)(?=\(|\{|$)(.*?)$/;

  $ChainLength = defined $ChainLength ? $ChainLength : '';
  $DoubleBondCount = defined $DoubleBondCount ? $DoubleBondCount : '';

  if (IsEmpty($ChainLength)) {
    return ($ChainLength, $DoubleBondCount, $DoubleBondGeometry, $Substituents, $Rings);
  }

  if ($ChainLength =~ /^O-/i) {
    $ChainLength =~ s/^O-//i;
  }
  elsif ($ChainLength =~ /^P-/i) {
    $ChainLength =~ s/^P-//i;
  }
  elsif ($ChainLength =~ /^d/i) {
    $ChainLength =~ s/^d//i;
  }
  elsif ($ChainLength =~ /^t/i) {
    $ChainLength =~ s/^t//i;
  }

  $ChainSpec = defined $LeftOverChainSpec ? $LeftOverChainSpec : ''; $LeftOverChainSpec = '';
  if ($ChainSpec && $DoubleBondCount) {
    ($DoubleBondGeometry, $LeftOverChainSpec) = $ChainSpec =~ /^\((.+?)\)(.*?)$/;
    $DoubleBondGeometry = defined $DoubleBondGeometry ? $DoubleBondGeometry : '';
    $ChainSpec = defined $LeftOverChainSpec ? $LeftOverChainSpec : ''; $LeftOverChainSpec = '';
  }
  if ($ChainSpec =~ /^\(/) {
    # Substituents...
    ($Substituents, $LeftOverChainSpec) = $ChainSpec =~ /^\((.+?)\)(.*?)$/;
    $Substituents = defined $Substituents ? $Substituents : '';
    $ChainSpec = defined $LeftOverChainSpec ? $LeftOverChainSpec : ''; $LeftOverChainSpec = '';
  }
  if ($ChainSpec && ($ChainSpec =~ /^\{/)) {
    # Rings...
    ($Rings, $LeftOverChainSpec) = $ChainSpec =~ /^\{(.+?)\}(.*?)$/;
    $Rings = defined $Rings ? $Rings : '';
    $ChainSpec = defined $LeftOverChainSpec ? $LeftOverChainSpec : ''; $LeftOverChainSpec = '';
  }
  return ($ChainLength, $DoubleBondCount, $DoubleBondGeometry, $Substituents, $Rings);
}


# Generate lipid ontology data...
sub GenerateLipidOntologyData {
  my($Molecule) = @_;
  my($Index, $DataFieldName, $DataFieldValue, $MethodName, @Atoms, @Bonds, @OntologyDataFieldValues);

  @OntologyDataFieldValues = ();

  @Atoms = $Molecule->GetAtoms();
  @Bonds = $Molecule->GetBonds();

  if (!@Atoms) {
    return \@OntologyDataFieldValues;
  }

  if (!$Molecule->DetectRings()) {
    return \@OntologyDataFieldValues;
  }

  for $Index (0 .. $#OntologyDataFieldNames) {
    $DataFieldName = $OntologyDataFieldNames[$Index];
    $DataFieldValue = 0;

    $MethodName = "GenerateOntologyDataFor${DataFieldName}";

    # strict refs is disabled temporarily to allow subroutine invocation
    # using string name...

    no strict "refs";
    $DataFieldValue = &${MethodName}($Molecule, \@Atoms, \@Bonds);
    use strict;

    push @OntologyDataFieldValues, $DataFieldValue;
  }

  @Atoms = ();
  @Bonds = ();

  return \@OntologyDataFieldValues;

}

# Generate fingerprints ontology data for fingerprins...
#
sub GenerateLipidOntologyDataForFingerprints {
  my($LipidOntologyDataRef) = @_;
  my(@LipidOntologyData);

  @LipidOntologyData = ();
  push @LipidOntologyData, @{$LipidOntologyDataRef};

  $FingerprintsString  = '';

  $FingerprintsString  = join "", map { $_ <= 99 ? sprintf "%02i", $_ : "00"; } @LipidOntologyData;
}

# Generate ontology value for CARBONS
#
sub GenerateOntologyDataForCARBONS {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for DOUBLEBONDS
#
# Notes:
#    . Number of double bonds between Carbons.
#
sub GenerateOntologyDataForDOUBLEBONDS {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom1, $Atom2, $Bond);

  $Value = 0;

  BOND: for $Bond (@{$BondsRef}) {
    ($Atom1, $Atom2) = $Bond->GetAtoms();
    if (!$Bond->IsDouble()) {
      next BOND;
    }
    if (!($Atom1->IsCarbon() && $Atom2->IsCarbon())) {
      next BOND;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for TRIPLEBONDS
#
# Notes:
#    . Number of triple bonds between Carbons
#
sub GenerateOntologyDataForTRIPLEBONDS {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom1, $Atom2, $Bond);

  $Value = 0;

  BOND: for $Bond (@{$BondsRef}) {
    ($Atom1, $Atom2) = $Bond->GetAtoms();
    if (!$Bond->IsTriple()) {
      next BOND;
    }
    if (!($Atom1->IsCarbon() && $Atom2->IsCarbon())) {
      next BOND;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for RINGS
#
sub GenerateOntologyDataForRINGS {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = $Molecule->GetNumOfRings();

  return $Value;
}

# Generate ontology value for NH2
#
sub GenerateOntologyDataForNH2 {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsNitrogen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('N.T3.H2', ['C'], ['-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for OH
#
sub GenerateOntologyDataForOH {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsOxygen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('O.T2.H1', ['C'], ['-'], ['!O'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for OOH
#
sub GenerateOntologyDataForOOH {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsOxygen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('O.X2', ['O.T2.H1', 'C'], ['-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for KETONES
#
sub GenerateOntologyDataForKETONES {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X3.DB1', ['O', 'C', 'C'], ['=', '-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for EPOXIDES
#
sub GenerateOntologyDataForEPOXIDES {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsOxygen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('O.X2.RA3', ['C.RA3', 'C.RA3'], ['-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for COOH
#
sub GenerateOntologyDataForCOOH {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.T3.DB1', ['O', 'O.T2.H1', 'C'], ['=', '-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for METHYLS
#
sub GenerateOntologyDataForMETHYLS {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.H3', ['C.X3,C.X4'], ['-'])) {
      next ATOM;
    }
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('C.X3,C.X4', $Atom)) {
      if ($AtomNeighbor1->DoesAtomNeighborhoodMatch('C.X4,C.X3.H1', ['C.X1.H3', 'C', 'C'], ['-', '-','-'])) {
	$Value++;
	next ATOM;
      }
      elsif ($AtomNeighbor1->DoesAtomNeighborhoodMatch('C.X3.DB1', ['C.X1.H3', 'C', 'C'], ['-', '=','-'])) {
	$Value++;
	next ATOM;
      }
    }
  }
  return $Value;
}

# Generate ontology value for SH
#
sub GenerateOntologyDataForSH {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsSulfur()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('S.T2.H1', ['C'], ['-'], ['!O'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for BR
#
sub GenerateOntologyDataForBR {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  for $Atom (@{$AtomsRef}) {
    if ($Atom->IsBromine()) {
      $Value++;
    }
  }

  return $Value;
}

# Generate ontology value for CL
#
sub GenerateOntologyDataForCL {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  for $Atom (@{$AtomsRef}) {
    if ($Atom->IsChlorine()) {
      $Value++;
    }
  }

  return $Value;
}

# Generate ontology value for F
#
sub GenerateOntologyDataForF {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  for $Atom (@{$AtomsRef}) {
    if ($Atom->IsFluorine()) {
      $Value++;
    }
  }

  return $Value;
}

# Generate ontology value for METHYLENES
#
sub GenerateOntologyDataForMETHYLENES {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.DB1', ['C'], ['='])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for CHO
#
sub GenerateOntologyDataForCHO {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X2.DB1.H1', ['O', 'C', 'H'], ['=', '-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for OME
#
sub GenerateOntologyDataForOME {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    # Is it a terminal Carbon attached to divalent Oxygen?
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.H3', ['O.X2'], ['-'])) {
      next ATOM;
    }
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('O.X2')) {
      if ($AtomNeighbor1->DoesAtomNeighborhoodMatch('O.X2', ['C.X3,C.X4', 'C.X1.H3'], ['-', '-'], ['C,H', 'C,H'])) {
	$Value++;
	next ATOM;
      }
    }
  }

  return $Value;
}

# Generate ontology value for OAC
#
sub GenerateOntologyDataForOAC {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1, $AtomNeighbor2);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    # Is it a terminal Carbon attached to Carbon with two Hydrogens?
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.H3', ['C.X2.H2'], ['-'])) {
      next ATOM;
    }
    # Go over the other Carbon atom neighbor...
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('C.X2', $Atom)) {
      # Is it attached to divalent Oxygen?
      if (!$AtomNeighbor1->DoesAtomNeighborhoodMatch('C.X2.H2', ['O.X2'], ['-'])) {
	next ATOM;
      }
      for $AtomNeighbor2 ($AtomNeighbor1->GetNeighborsUsingAtomSpecification('O.X2')) {
	if ($AtomNeighbor2->DoesAtomNeighborhoodMatch('O.X2', ['C.X3,C.X4', 'C.X2.H2'], ['-', '-'], ['C,H', 'C,H'])) {
	  $Value++;
	  next ATOM;
	}
      }
    }
  }

  return $Value;
}

# Generate ontology value for ET
#
sub GenerateOntologyDataForET {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1, $AtomNeighbor2);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.H3', ['C.X2.H2'], ['-'])) {
      next ATOM;
    }
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('C.X2.H2', $Atom)) {
      if (!$AtomNeighbor1->DoesAtomNeighborhoodMatch('C.X2.H2', ['C.X3,C.X4'], ['-'])) {
	next ATOM;
      }
      for $AtomNeighbor2 ($AtomNeighbor1->GetNeighborsUsingAtomSpecification('C.X3,C.X4', $Atom, $AtomNeighbor1)) {
	if ($AtomNeighbor2->DoesAtomNeighborhoodMatch('C.X4,C.X3.H1', ['C.X2.H2', 'C', 'C'], ['-', '-','-'])) {
	  $Value++;
	  next ATOM;
	}
	elsif ($AtomNeighbor2->DoesAtomNeighborhoodMatch('C.X3.DB1', ['C.X2.H2', 'C', 'C'], ['-', '=','-'])) {
	  $Value++;
	  next ATOM;
	}
      }
    }
  }
  return $Value;
}

# Generate ontology value for PR
#
sub GenerateOntologyDataForPR {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1, $AtomNeighbor2, $AtomNeighbor3);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1.H3', ['C.X2.H2'], ['-'])) {
      next ATOM;
    }
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('C.X2.H2', $Atom)) {
      if (!$AtomNeighbor1->DoesAtomNeighborhoodMatch('C.X2.H2', ['C.X2.H2'], ['-'])) {
	next ATOM;
      }
      for $AtomNeighbor2 ($AtomNeighbor1->GetNeighborsUsingAtomSpecification('C.X2.H2', $Atom, $AtomNeighbor1)) {
	if (!$AtomNeighbor2->DoesAtomNeighborhoodMatch('C.X2.H2', ['C.X3,C.X4'], ['-'])) {
	  next ATOM;
	}
	for $AtomNeighbor3 ($AtomNeighbor2->GetNeighborsUsingAtomSpecification('C.X3,C.X4', $Atom, $AtomNeighbor1, $AtomNeighbor2)) {
	  if ($AtomNeighbor3->DoesAtomNeighborhoodMatch('C.X4,C.X3.H1', ['C.X2.H2', 'C', 'C'], ['-', '-','-'])) {
	    $Value++;
	    next ATOM;
	  }
	  elsif ($AtomNeighbor3->DoesAtomNeighborhoodMatch('C.X3.DB1', ['C.X2.H2', 'C', 'C'], ['-', '=','-'])) {
	    $Value++;
	    next ATOM;
	  }
	}
      }
    }
  }

  return $Value;
}

# Generate ontology value for COOME
#
sub GenerateOntologyDataForCOOME {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom, $AtomNeighbor1, $AtomNeighbor2);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    # Is it a terminal Carbon attached to divalent Oxygen?
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X1', ['O.X2'], ['-'])) {
      next ATOM;
    }
    for $AtomNeighbor1 ($Atom->GetNeighborsUsingAtomSpecification('O.X2')) {
      for $AtomNeighbor2 ($AtomNeighbor1->GetNeighborsUsingAtomSpecification('C.X3.DB1')) {
	if ($AtomNeighbor2->DoesAtomNeighborhoodMatch('C.X3.DB1', ['O', 'O.X2', 'C'], ['=', '-', '-'], ['C', 'C', 'C,H'])) {
	  $Value++;
	  next ATOM;
	}
      }
    }
  }

  return $Value;
}

# Generate ontology value for NO2
#
sub GenerateOntologyDataForNO2 {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsNitrogen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('N.DB1.FC+1', ['O', 'O.X1.FC-1', '!O'], ['=', '-', '-'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for ESTER
#
sub GenerateOntologyDataForESTER {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X3.DB1', ['O', 'O.X2', 'C'], ['=', '-', '-'], ['C', 'C', 'C,H'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for ETHER
#
sub GenerateOntologyDataForETHER {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsOxygen()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('O.X2', ['C.X2.H2', 'C.X2.H2'], ['-', '-'], ['C,H', 'C,H'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Generate ontology value for AMIDES
#
sub GenerateOntologyDataForAMIDES {
  my($Molecule, $AtomsRef, $BondsRef) = @_;
  my($Value, $Atom);

  $Value = 0;

  ATOM: for $Atom (@{$AtomsRef}) {
    if (!$Atom->IsCarbon()) {
      next ATOM;
    }
    if (!$Atom->DoesAtomNeighborhoodMatch('C.X3.DB1', ['O', 'N.T3', 'C'], ['=', '-', '-'], ['C', 'C,H', 'C,H'])) {
      next ATOM;
    }
    $Value++;
  }

  return $Value;
}

# Retrieve information about input SD files...
sub RetrieveSDFilesInfo {
  my($SDFile, $Index, $FileDir, $FileExt, $FileName, $OutFileRoot, $SDOutFileExt, $NewSDFileName, $TextOutFileExt, $NewTextFileName);

  %SDFilesInfo = ();

  @{$SDFilesInfo{FileOkay}} = ();
  @{$SDFilesInfo{OutFileRoot}} = ();
  @{$SDFilesInfo{SDOutFileNames}} = ();
  @{$SDFilesInfo{TextOutFileNames}} = ();


 FILELIST: for $Index (0 .. $#SDFilesList) {
    $SDFile = $SDFilesList[$Index];
    $SDFilesInfo{FileOkay}[$Index] = 0;

    if (!(-e $SDFile)) {
      warn "Warning: Ignoring file $SDFile: It doesn't exist\n";
      next FILELIST;
    }

    # Check data field name in SD file...
    if ($CompoundIDMode =~ /^DataField$/i) {
      my($CmpdString, $SpecifiedDataField, @CmpdLines, %DataFieldValues);

      $SpecifiedDataField = $CompoundID;

      @CmpdLines = ();
      open SDFILE, "$SDFile" or die "Error: Couldn't open $SDFile: $! \n";
      $CmpdString = ReadCmpdString(\*SDFILE);
      close SDFILE;
      @CmpdLines = split "\n", $CmpdString;
      %DataFieldValues = GetCmpdDataHeaderLabelsAndValues(\@CmpdLines);

      if (!exists $DataFieldValues{$SpecifiedDataField}) {
	warn "Warning: Ignoring file $SDFile: Data field name, $SpecifiedDataField, specified using  \"--CompoundID\" option for \"DataField\" value of \"--CompoundIDMode\" doesn't exist\n";
	next FILELIST;
      }
    }

    # Setup output file names...
    $FileDir = ""; $FileName = ""; $FileExt = "";
    ($FileDir, $FileName, $FileExt) = ParseFileName($SDFile);

    $TextOutFileExt = "csv";
    if ($Options{outdelim} =~ /^tab$/i) {
      $TextOutFileExt = "tsv";
    }
    $SDOutFileExt = $FileExt;

    if ($OutFileRoot && (@SDFilesList == 1)) {
      my ($RootFileDir, $RootFileName, $RootFileExt) = ParseFileName($OutFileRoot);
      if ($RootFileName && $RootFileExt) {
	$FileName = $RootFileName;
      }
      else {
	$FileName = $OutFileRoot;
      }
      $OutFileRoot = $FileName;
    }
    else {
      $OutFileRoot = "${FileName}OntologyData";
    }

    $NewSDFileName = "${OutFileRoot}.${SDOutFileExt}";
    $NewTextFileName = "${OutFileRoot}.${TextOutFileExt}";

    if ($SDOutput) {
      if ($SDFile =~ /$NewSDFileName/i) {
	warn "Warning: Ignoring input file $SDFile: Same output, $NewSDFileName, and input file names.\n";
	print "Specify a different name using \"-r --root\" option or use default name.\n";
	next FILELIST;
      }
    }

    if (!$OverwriteFiles) {
      if ($SDOutput) {
	if (-e $NewSDFileName) {
	  warn "Warning: Ignoring file $SDFile: The file $NewSDFileName already exists\n";
	  next FILELIST;
	}
      }
      if ($TextOutput) {
	if (-e $NewTextFileName) {
	  warn "Warning: Ignoring file $SDFile: The file $NewTextFileName already exists\n";
	  next FILELIST;
	}
      }
    }

    $SDFilesInfo{FileOkay}[$Index] = 1;

    $SDFilesInfo{OutFileRoot}[$Index] = $OutFileRoot;
    $SDFilesInfo{SDOutFileNames}[$Index] = $NewSDFileName;
    $SDFilesInfo{TextOutFileNames}[$Index] = $NewTextFileName;
  }
}

# Process options...
sub ProcessOptions {

  # Intialize fingerprints generation...
  $FingerprintsString = '';

  # Supported ontology data field names...
  @OntologyDataFieldNames = qw(CARBONS DOUBLEBONDS TRIPLEBONDS RINGS NH2 OH OOH KETONES EPOXIDES COOH METHYLS SH BR CL F METHYLENES CHO OME OAC ET PR COOME NO2 ESTER ETHER AMIDES);

  # Ontology data field name to index map...
  my($DataFieldIndex, $DataFieldName);
  %OntologyDataFieldNameToIndexMap = ();
  for $DataFieldIndex (0 .. $#OntologyDataFieldNames) {
    $DataFieldName = $OntologyDataFieldNames[$DataFieldIndex];
    $OntologyDataFieldNameToIndexMap{$DataFieldName} = $DataFieldIndex;
  }

  $Mode = $Options{mode};

  $CompoundIDMode = $Options{compoundidmode};
  if (IsEmpty($CompoundIDMode)) {
    $CompoundIDMode = $Options{mode} =~ /^LMSD$/i ? 'DataField' : 'MolNameOrLabelPrefix';
  }

  $CompoundID = $Options{compoundid};
  if (IsEmpty($CompoundID)) {
    $CompoundID = $Options{mode} =~ /^LMSD$/i ? 'LM_ID' : 'Cmpd';
  }

  $CompoundIDLabel = $Options{compoundidlabel};
  if (IsEmpty($CompoundIDLabel)) {
    $CompoundIDLabel = $Options{mode} =~ /^LMSD$/i ? $CompoundID : 'CompoundID';
  }

  $ProcessLMIDValue = ( $Mode =~ /^LMSD$/i && $CompoundIDMode =~ /^DataField$/i && $CompoundID =~ /^LM_ID$/i) ? 1 : 0;

  $Fingerprints = $Options{fingerprint};
  if (IsEmpty($Fingerprints)) {
    $Fingerprints = $Options{mode} =~ /^LMSD$/i ? 'Yes' : 'No';
  }
  $Fingerprints = ($Fingerprints =~ /^Yes$/i) ? 1 : 0;

  $Output = $Options{output};
  $SDOutput = ($Options{output} =~ /^(SD|Both)$/i) ? 1 : 0;
  $TextOutput = ($Options{output} =~ /^(Text|Both)$/i) ? 1 : 0;

  $OutDelim = ($Options{outdelim} =~ /tab/i ) ? "\t" : (($Options{outdelim} =~ /semicolon/i) ? "\;" : "\,");
  $OutQuote = ($Options{quote} =~ /^Yes$/i) ? 1 : 0;

  $OverwriteFiles = $Options{overwrite} ? 1 : 0;
  $OutFileRoot = $Options{root} ? $Options{root} : 0;

}


# Setup script usage  and retrieve command line arguments specified using various options...
sub SetupScriptUsage {

  # Setup default and retrieve all the options...
  %Options = ();

  $Options{compoundid} = '';
  $Options{compoundidlabel} = '';
  $Options{compoundidmode} = '';

  $Options{fingerprint} = '';

  $Options{mode} = 'LMSD';

  $Options{output} = 'text';
  $Options{outdelim} = 'comma';
  $Options{quote} = 'yes';

  GetOptions(\%Options, "compoundid=s", "compoundidlabel=s", "compoundidmode=s", "fingerprint|f=s", "mode|m=s", "outdelim=s", "output=s", "overwrite|o", "quote|q=s", "root|r=s", "workingdir|w=s");
  if ($Options{workingdir}) {
    if (! -d $Options{workingdir}) {
      die "Error: The value specified, $Options{workingdir}, for option \"-w --workingdir\" is not a directory name.\n";
    }
    chdir $Options{workingdir} || die "Error: Couldn't chdir $Options{workingdir}: $! \n";
  }
  if ($Options{fingerprint} && $Options{fingerprint} !~ /^(Yes|No)$/i) {
    die "Error: The value specified, $Options{fingerprint}, for option \"-f --fingerprint\" is not valid. Allowed values: Yes or No\n";
  }
  if ($Options{compoundidmode} && $Options{compoundidmode} !~ /^(DataField|MolName|LabelPrefix|MolNameOrLabelPrefix)$/i) {
    die "Error: The value specified, $Options{compoundidmode}, for option \"--CompoundIDMode\" is not valid. Allowed values: DataField, MolName, LabelPrefix, or MolNameOrLabelPrefix\n";
  }
  if ($Options{mode} !~ /^(LMSD|Generic)$/i) {
    die "Error: The value specified, $Options{mode}, for option \"-m, --mode\" is not valid. Allowed values: LMSD, or Generic\n";
  }
  if ($Options{output} !~ /^(SD|text|both)$/i) {
    die "Error: The value specified, $Options{output}, for option \"--output\" is not valid. Allowed values: SD, text, or both\n";
  }
  if ($Options{outdelim} !~ /^(comma|semicolon|tab)$/i) {
    die "Error: The value specified, $Options{outdelim}, for option \"--outdelim\" is not valid. Allowed values: comma, tab, or semicolon\n";
  }
  if ($Options{quote} !~ /^(Yes|No)$/i) {
    die "Error: The value specified, $Options{quote}, for option \"-q --quote\" is not valid. Allowed values: Yes or No\n";
  }
  if ($Options{outdelim} =~ /semicolon/i && $Options{quote} =~ /^No$/i) {
    die "Error: The value specified, $Options{quote}, for option \"-q --quote\" is not allowed with, semicolon value of \"--outdelim\" option: Fingerprints string use semicolon as delimiter for various data fields and must be quoted.\n";
  }
}

__END__

=head1 NAME

GenerateLMOntologyData.pl - Generate lipid ontology data for lipids in SDFile(s)

=head1 SYNOPSIS

GenerateLMOntologyData.pl SDFile(s)...

GenerateLMOntologyData.pl [B<--CompoundID> DataFieldName or LabelPrefixString]
[B<--CompoundIDLabel> text] [B<--CompoundIDMode> DataField | MolName | LabelPrefix | MolNameOrLabelPrefix]
[B<-f, --fingerprint> Yes | No] [B<-h, --help>] [B<-m, --mode> LMSD | Generic]
[B<--OutDelim> I<comma | tab | semicolon>] [B<--output> I<SD | text | both>] [B<-o, --overwrite>]
[B<-q, --quote> I<Yes | No>] [B<-r, --root> I<RootName>]
[B<-w,--WorkingDir> dirname] SDFile(s)...

=head1 DESCRIPTION

Generate LIPID MAPS ontology data using structure data in SD file(s). The text/ontology
search available on LIPID MAPS website provides support for using the ontology data during
database searching. The URL is:

http://www.lipidmaps.org/data/structure/LMSDSearch.php?Mode=SetupTextOntologySearch

The current release of ontology data generation script generates the count of following
26 molecular descriptors:

    Abbreviation    Description

    CARBONS         # of carbons
    DOUBLEBONDS     # of double bonds
    TRIPLEBONDS     # of triple bonds
    RINGS           # of rings
    NH2             # of amino groups
    OH              # of hydroxy groups
    OOH             # of hydroperoxy groups
    KETONES         # of keto groups
    EPOXIDES        # of of epoxy groups
    COOH            # of carboxy groups
    METHYLS         # of methyl groups
    SH              # of thio groups
    BR              # of bromo groups
    CL              # of chloro groups
    F               # of fluoro groups
    METHYLENES      # of methylene groups
    CHO             # of aldehyde groups
    OME             # of methoxy groups
    OAC             # of acetoxy groups
    ET              # of ethyl groups
    PR              # of propyl groups
    COOME           # methyl ester groups
    NO2             # of nitro groups
    ESTER           # of ester groups
    ETHER           # of ether groups
    AMIDES          # of amide groups

Additionally, a  52 character FINGERPRINT string, with two characters corresponding to value
of each ontology data type, can also be generated.

=head1 OPTIONS

=over 4

=item B<--CompoundID> I<DataFieldName or LabelPrefixString>

This value is B<--CompoundIDMode> specific and indicates how compound ID is generated.

For I<DataField> value of B<--CompoundIDMode> option, it corresponds to datafield label name
whose value is used as compound ID; otherwise, it's a prefix string used for generating compound
IDs like LabelPrefixString<Number>.

Default value, I<Cmpd>, during I<LabelPrefix> or I<MolNameOrLabelPrefix> values of
B<--CompoundIDMode> option generates compound IDs which look like Cmpd<Number>.

Default value, I<LM_ID>, during I<DataField> value of B<--CompoundIDMode> option
uses value of I<LM_ID> data field in I<SDFile(s)> as compound ID.

Examples for I<DataField> value of B<--CompoundIDMode>:

    Reg_No
    LM_ID

Examples for I<LabelPrefix> or I<MolNameOrLabelPrefix> value of B<--CompoundIDMode>:

    Compound

The value specified above generates compound IDs which correspond to Compound<Number>
instead of default value of Cmpd<Number>.

=item B<--CompoundIDLabel> I<text>

Specify compound ID column label for CSV/TSV text file(s). Default value: I<CompoundID>.

=item B<--CompoundIDMode> I<DataField | MolName | LabelPrefix | MolNameOrLabelPrefix>

Specify how to generate compound IDs and write to CSV/TSV text file(s) along with calculated
ontology values for I<text | both> values of B<--output> option: use a I<SDFile(s)>
datafield value; use molname line from I<SDFile(s)>; generate a sequential ID with specific prefix;
use combination of both MolName and LabelPrefix with usage of LabelPrefix values for empty
molname lines.

Possible values: I<DataField | MolName | LabelPrefix | MolNameOrLabelPrefix>.
Default value: I<DataField> during I<LMSD> value of <-m, --mode> option; I<LabelPrefix> during
I<Generic> value of <-m, --mode> option.

For I<MolNameAndLabelPrefix> value of B<--CompoundIDMode>, molname line in I<SDFile(s)> takes
precedence over sequential compound IDs generated using I<LabelPrefix> and only empty molname
values are replaced with sequential compound IDs.

=item B<-f, --fingerprint> I<Yes | No>

Specify whether to generate a 52 fingerprints string, with two characters corresponding
to value of each ontology data type. Possible values: I<Yes or No>.

Default value: I<Yes> during I<LMSD> value of <-m, --mode> option; I<No> during
I<Generic> value of <-m, --mode> option.

=item B<-h, --help>

Print this help message

=item B<-m, --mode> I<LMSD | Generic>

Specify whether input I<SDFile(s)> contain lipids structure data corresponding to to LIPID
MAPS LMSD database or any arbitrary set of compounds. Possible values: I<LMSD or Generic>.
Default: I<LMSD>.

During I<LMSD> value of B<-m, --mode> option, data field B<LM_ID> must be present in
I<SDFile(s)> to generate ontology data. Based on the presence of a valid LIPID MAPS
abbreviation as value of B<NAME> data field in I<SDFile(s)>, number of carbons ontology
value for glycerolipids, glycerophospholipids and sphingolipids is adjusted to carbons in chains
instead of total number of carbons.

For I<Generic> value of B<-m, --mode> option, no adjustments are made to the value of
number of carbons and it corresponds to the total number of carbons in a compound.

=item B<--OutDelim> I<comma | tab | semicolon>

Delimiter for output CSV/TSV text file(s). Possible values: I<comma, tab, or semicolon>
Default value: I<comma>

=item B<--output> I<SD | text | both>

Type of output files to generate. Possible values: I<SD, text, or both>. Default value: I<text>.

=item B<-o, --overwrite>

Overwrite existing files.

=item B<-q, --quote> I<Yes | No>

Put quote around column values in output CSV/TSV text file(s). Possible values:
I<Yes or No>. Default value: I<Yes>.

=item B<-r, --root> I<RootName>

New file name is generated using the root: <Root>.<Ext>. Default for new file names:
<SDFileName><OntologyData>.<Ext>. The file type determines <Ext> value.
The sdf, csv, and tsv <Ext> values are used for SD, comma/semicolon, and tab
delimited text files, respectively.This option is ignored for multiple input files.

=item B<-w,--WorkingDir> I<dirname>

Location of working directory. Default: current directory

=head1 EXAMPLES

To generate ontology data including fingerprints string for LMSD structure data
file containing LM_ID and generate SampleOntologyData.csv with LM_ID as
compound IDs along with ontology data, type:

    % GenerateLMOntologyData.pl -o Sample.sdf

To generate ontology data excluding fingerprints string for LMSD structure data
file containing LM_ID and generate SampleOntologyData.csv with LM_ID as
compound IDs along with ontology data, type:

    % GenerateLMOntologyData.pl -m LMSD --fingerprint No -o Sample.sdf

To generate ontology data excluding fingerprints string for any structure data
file and generate SampleOntologyData.csv with molname or sequential compound IDs
along with ontology data, type:

    % GenerateLMOntologyData.pl -m Generic -o Sample.sdf

To generate ontology data excluding fingerprints string for any structure data
file and generate SampleOntologyData.csv with sequential compound IDs
along with ontology data, type:

    % GenerateLMOntologyData.pl -m Generic --CompoundIDMode LabelPrefix
      -o Sample.sdf

To generate ontology data including fingerprints string for any structure data
file and generate both SampleOntologyData.csv and  SampleOntologyData.sdf files
with value of a data field as compound IDs in text file along with fingerprints and
ontology data, type:

    % GenerateLMOntologyData.pl -m Generic --CompoundIDMode DataField
      --CompoundID MOL_ID --fingerprints yes --output both -o Sample.sdf

To generate ontology data including fingerprints string for LMSD structure data
file containing LM_ID and generate SampleOntologyData.tsv with LM_ID as
compound IDs along with ontology data, type:

    % GenerateLMOntologyData.pl --OutDelim tab -o Sample.sdf

=head1 AUTHOR

Manish Sud

=head1 COPYRIGHT

Copyright (C) 2006-2012. The Regents of the University of California. All Rights Reserved.

=head1 LICENSE

Modified BSD License

=cut
